@@ -18,9 +18,11 @@
 #define R_NO_REMAP
 #include <R.h>
 #include <Rinternals.h>
+#include <Rversion.h>
 
+#include "array.h"
 #include "nanoarrow.h"
-
+#include "nanoarrow/r.h"
 #include "util.h"
 
 // Needed for the list_of materializer
@@ -56,10 +58,14 @@ SEXP nanoarrow_alloc_type(enum VectorType vector_type, R_xlen_t len) {
 // A version of Rf_getAttrib(x, sym) != R_NilValue that never
 // expands the row.names attribute
 static int has_attrib_safe(SEXP x, SEXP sym) {
+#if R_VERSION >= R_Version(4, 6, 0)
+  return (int)R_hasAttrib(x, sym);
+#else
   for (SEXP atts = ATTRIB(x); atts != R_NilValue; atts = CDR(atts)) {
     if (TAG(atts) == sym) return TRUE;
   }
   return FALSE;
+#endif
 }
 
 R_xlen_t nanoarrow_data_frame_size(SEXP x) {
@@ -108,10 +114,19 @@ int nanoarrow_ptype_is_data_frame(SEXP ptype) {
           (Rf_xlength(ptype) > 0 && has_attrib_safe(ptype, R_NamesSymbol)));
 }
 
+int nanoarrow_ptype_is_nanoarrow_vctr(SEXP ptype) {
+  return Rf_inherits(ptype, "nanoarrow_vctr");
+}
+
 SEXP nanoarrow_materialize_realloc(SEXP ptype, R_xlen_t len) {
   SEXP result;
 
-  if (Rf_isObject(ptype)) {
+  if (Rf_isMatrix(ptype)) {
+    // The actual value is built in the child converter but we can't have
+    // a NULL here because that confuses the internals into thinking that
+    // the allocate was never called.
+    result = PROTECT(Rf_allocVector(TYPEOF(ptype), 0));
+  } else if (Rf_isObject(ptype)) {
     // There may be a more accurate test that more precisely captures the case
     // where a user has specified a valid ptype that doesn't work in a preallocate
     // + fill conversion.
@@ -122,7 +137,27 @@ SEXP nanoarrow_materialize_realloc(SEXP ptype, R_xlen_t len) {
       }
     }
 
-    if (nanoarrow_ptype_is_data_frame(ptype)) {
+    if (nanoarrow_ptype_is_nanoarrow_vctr(ptype)) {
+      // The object we return here is one that will accumulate chunks and
+      // be finalized with a value (rather than being strictly copied into
+      // after every new chunk is seen).
+      result = PROTECT(Rf_allocVector(INTSXP, len));
+      Rf_copyMostAttrib(ptype, result);
+
+      // For the purposes of building the list of chunks, chunks is a pairlist
+      // (it will be converted to a regular list when this converter is finalized)
+      // Technically the first value here won't be used (this simplifies the
+      // appending).
+      SEXP chunks_list = PROTECT(Rf_list1(R_NilValue));
+
+      // To start, the chunks list and the end of the chunks list are the same node
+      SEXP chunks_tail_sym = PROTECT(Rf_install("chunks_tail"));
+      SEXP chunks_sym = PROTECT(Rf_install("chunks"));
+      Rf_setAttrib(result, chunks_sym, chunks_list);
+      Rf_setAttrib(result, chunks_tail_sym, chunks_list);
+
+      UNPROTECT(3);
+    } else if (nanoarrow_ptype_is_data_frame(ptype)) {
       R_xlen_t num_cols = Rf_xlength(ptype);
       result = PROTECT(Rf_allocVector(VECSXP, num_cols));
       for (R_xlen_t i = 0; i < num_cols; i++) {
@@ -207,6 +242,10 @@ static void fill_vec_with_nulls(SEXP x, R_xlen_t offset, R_xlen_t len) {
 }
 
 static void copy_vec_into(SEXP x, SEXP dst, R_xlen_t offset, R_xlen_t len) {
+  if (nanoarrow_ptype_is_nanoarrow_vctr(dst)) {
+    Rf_error("Can't copy_vec_into() to nanoarrow_vctr");
+  }
+
   if (nanoarrow_ptype_is_data_frame(dst)) {
     if (!nanoarrow_ptype_is_data_frame(x)) {
       Rf_error("Expected record-style vctr result but got non-record-style result");
@@ -270,6 +309,124 @@ static void copy_vec_into(SEXP x, SEXP dst, R_xlen_t offset, R_xlen_t len) {
   }
 }
 
+int nanoarrow_materialize_finalize_result(SEXP converter_xptr) {
+  SEXP converter_shelter = R_ExternalPtrProtected(converter_xptr);
+  struct RConverter* converter = (struct RConverter*)R_ExternalPtrAddr(converter_xptr);
+  SEXP result = VECTOR_ELT(converter_shelter, 4);
+
+  // Materialize never called (e.g., empty stream)
+  if (result == R_NilValue) {
+    nanoarrow_converter_reserve(converter_xptr, 0);
+    result = VECTOR_ELT(converter_shelter, 4);
+  }
+
+  if (nanoarrow_ptype_is_nanoarrow_vctr(result)) {
+    // Get the schema for this converter. Technically this will overwrite
+    // a schema that was provided explicitly; however, we currently do not
+    // handle that case.
+
+    SEXP schema_xptr = VECTOR_ELT(converter_shelter, 1);
+
+    // We no longer need to keep track of chunks_tail
+    SEXP chunks_tail_sym = PROTECT(Rf_install("chunks_tail"));
+    Rf_setAttrib(result, chunks_tail_sym, R_NilValue);
+
+    // We also want to pass on the class of the ptype we received
+    SEXP subclass_sexp = Rf_getAttrib(result, R_ClassSymbol);
+
+    // We no longer need the first element of the pairlist, which was
+    // intentionally set to R_NilValue.
+    SEXP chunks_sym = PROTECT(Rf_install("chunks"));
+    SEXP chunks_pairlist0 = Rf_getAttrib(result, chunks_sym);
+
+    // If there were zero chunks, there will be no "first" node
+    SEXP chunks_list;
+    if (CDR(chunks_pairlist0) == R_NilValue) {
+      chunks_list = PROTECT(Rf_allocVector(VECSXP, 0));
+    } else {
+      chunks_list = PROTECT(Rf_PairToVectorList(CDR(chunks_pairlist0)));
+    }
+
+    // Set up the call to new_nanoarrow_vctr
+    SEXP new_nanoarrow_vctr_sym = PROTECT(Rf_install("new_nanoarrow_vctr"));
+    SEXP new_nanoarrow_vctr_call = PROTECT(
+        Rf_lang4(new_nanoarrow_vctr_sym, chunks_list, schema_xptr, subclass_sexp));
+    SEXP final_result = PROTECT(Rf_eval(new_nanoarrow_vctr_call, nanoarrow_ns_pkg));
+
+    SET_VECTOR_ELT(converter_shelter, 4, final_result);
+    UNPROTECT(6);
+  } else if (nanoarrow_ptype_is_data_frame(result)) {
+    // For each child, finalize the result and then reassign it
+    SEXP child_converter_xptrs = VECTOR_ELT(converter_shelter, 3);
+    for (R_xlen_t i = 0; i < Rf_xlength(child_converter_xptrs); i++) {
+      SEXP child_converter_xptr = VECTOR_ELT(child_converter_xptrs, i);
+      NANOARROW_RETURN_NOT_OK(
+          nanoarrow_materialize_finalize_result(child_converter_xptr));
+
+      SEXP child_result =
+          PROTECT(nanoarrow_converter_release_result(child_converter_xptr));
+      SET_VECTOR_ELT(result, i, child_result);
+      UNPROTECT(1);
+    }
+  } else if (converter->ptype_view.vector_type == VECTOR_TYPE_MATRIX) {
+    SEXP child_converter_xptrs = VECTOR_ELT(converter_shelter, 3);
+    SEXP item_converter_xptr = VECTOR_ELT(child_converter_xptrs, 0);
+    NANOARROW_RETURN_NOT_OK(nanoarrow_materialize_finalize_result(item_converter_xptr));
+    SEXP item_result = PROTECT(nanoarrow_converter_release_result(item_converter_xptr));
+
+    SEXP matrix_symbol = PROTECT(Rf_install("matrix"));
+    SEXP nrow_sexp = PROTECT(
+        Rf_ScalarInteger(Rf_xlength(item_result) / converter->schema_view.fixed_size));
+    SEXP ncol_sexp = PROTECT(Rf_ScalarInteger(converter->schema_view.fixed_size));
+    SEXP byrow_sexp = PROTECT(Rf_ScalarLogical(TRUE));
+    SEXP matrix_call =
+        PROTECT(Rf_lang5(matrix_symbol, item_result, nrow_sexp, ncol_sexp, byrow_sexp));
+    SEXP final_result = PROTECT(Rf_eval(matrix_call, R_BaseNamespace));
+    SET_VECTOR_ELT(converter_shelter, 4, final_result);
+    UNPROTECT(7);
+  }
+
+  return NANOARROW_OK;
+}
+
+static int nanoarrow_materialize_nanoarrow_vctr(struct RConverter* converter,
+                                                SEXP converter_xptr) {
+  // This is a case where the callee needs ownership, which we can do via a
+  // shallow copy.
+  SEXP converter_shelter = R_ExternalPtrProtected(converter_xptr);
+  SEXP schema_xptr = VECTOR_ELT(converter_shelter, 1);
+  SEXP array_xptr = VECTOR_ELT(converter_shelter, 2);
+
+  SEXP array_out_xptr = PROTECT(nanoarrow_array_owning_xptr());
+  struct ArrowArray* out_array = nanoarrow_output_array_from_xptr(array_out_xptr);
+  array_export(array_xptr, out_array);
+  R_SetExternalPtrTag(array_out_xptr, schema_xptr);
+
+  // Update the offset/length in case a slice is being requested from the
+  // converter.
+  out_array->offset += converter->src.offset;
+  out_array->length = converter->src.length;
+
+  // Get the cached copy of the pairlist node at the end of the current
+  // chunks list.
+  SEXP chunks_tail_sym = PROTECT(Rf_install("chunks_tail"));
+  SEXP chunks_tail = PROTECT(Rf_getAttrib(converter->dst.vec_sexp, chunks_tail_sym));
+
+  // Create a length-1 pairlist node containing the chunk
+  SEXP next_sexp = PROTECT(Rf_cons(array_out_xptr, R_NilValue));
+
+  // Append it to the end of the current pairlist
+  SETCDR(chunks_tail, next_sexp);
+  UNPROTECT(1);
+
+  // Update the cached copy of the pairlist node at the end of the current
+  // chunks list.
+  Rf_setAttrib(converter->dst.vec_sexp, chunks_tail_sym, next_sexp);
+  UNPROTECT(3);
+
+  return NANOARROW_OK;
+}
+
 static int nanoarrow_materialize_other(struct RConverter* converter,
                                        SEXP converter_xptr) {
   // Ensure that we have a ptype SEXP to send in the call back to R
@@ -280,20 +437,19 @@ static int nanoarrow_materialize_other(struct RConverter* converter,
     UNPROTECT(1);
   }
 
-  // A unique situation where we don't want owning external pointers because we know
-  // these are protected for the duration of our call into R and because we don't want
-  // the underlying array to be released and invalidate the converter. The R code in
-  // convert_fallback_other() takes care of ensuring an independent copy with the correct
-  // offset/length.
-  SEXP schema_xptr = PROTECT(R_MakeExternalPtr(
-      (struct ArrowSchema*)converter->schema_view.schema, R_NilValue, R_NilValue));
-  Rf_setAttrib(schema_xptr, R_ClassSymbol, nanoarrow_cls_schema);
-  // We do need to set the protected member of the array external pointer to signal that
-  // it is not an independent array (i.e., force a shallow copy).
-  SEXP array_xptr = PROTECT(R_MakeExternalPtr(
-      (struct ArrowArray*)converter->array_view.array, schema_xptr, converter_xptr));
-  Rf_setAttrib(array_xptr, R_ClassSymbol, nanoarrow_cls_array);
+  // Special-case the nanoarrow_vctr conversion
+  if (Rf_inherits(converter->dst.vec_sexp, "nanoarrow_vctr")) {
+    return nanoarrow_materialize_nanoarrow_vctr(converter, converter_xptr);
+  }
 
+  // We've ensured proper ownership of array_xptr and ensured that its
+  // schema is set, so we can pass these safely to the R-level
+  // convert_fallback_other.
+  SEXP converter_shelter = R_ExternalPtrProtected(converter_xptr);
+  SEXP array_xptr = VECTOR_ELT(converter_shelter, 2);
+
+  // The R code in convert_fallback_other() takes care of ensuring an independent copy
+  // with the correct offset/length if it is necessary to update them.
   SEXP offset_sexp = PROTECT(
       Rf_ScalarReal((double)(converter->src.array_view->offset + converter->src.offset)));
   SEXP length_sexp = PROTECT(Rf_ScalarReal((double)converter->src.length));
@@ -307,7 +463,7 @@ static int nanoarrow_materialize_other(struct RConverter* converter,
   copy_vec_into(result_src, converter->dst.vec_sexp, converter->dst.offset,
                 converter->dst.length);
 
-  UNPROTECT(7);
+  UNPROTECT(5);
   return NANOARROW_OK;
 }
 
@@ -367,9 +523,7 @@ static int nanoarrow_materialize_data_frame(struct RConverter* converter,
 
 static int materialize_list_element(struct RConverter* converter, SEXP converter_xptr,
                                     int64_t offset, int64_t length) {
-  if (nanoarrow_converter_reserve(converter_xptr, length) != NANOARROW_OK) {
-    nanoarrow_converter_stop(converter_xptr);
-  }
+  nanoarrow_converter_reserve(converter_xptr, length);
 
   converter->src.offset = offset;
   converter->src.length = length;
@@ -452,6 +606,62 @@ static int nanoarrow_materialize_list_of(struct RConverter* converter,
   return NANOARROW_OK;
 }
 
+static int nanoarrow_materialize_matrix(struct RConverter* converter,
+                                        SEXP converter_xptr) {
+  SEXP converter_shelter = R_ExternalPtrProtected(converter_xptr);
+  SEXP child_converter_xptrs = VECTOR_ELT(converter_shelter, 3);
+  struct RConverter* child_converter = converter->children[0];
+  SEXP child_converter_xptr = VECTOR_ELT(child_converter_xptrs, 0);
+
+  struct ArrayViewSlice* src = &converter->src;
+
+  // Make sure we error for dictionary types
+  if (src->array_view->array->dictionary != NULL) {
+    return EINVAL;
+  }
+
+  switch (src->array_view->storage_type) {
+    case NANOARROW_TYPE_FIXED_SIZE_LIST:
+      break;
+    default:
+      return EINVAL;
+  }
+
+  int64_t raw_src_offset = src->array_view->offset + src->offset;
+  int64_t list_length = src->array_view->layout.child_size_elements;
+  int64_t child_length = list_length * src->length;
+
+  if (list_length != Rf_ncols(converter->ptype_view.ptype)) {
+    Rf_error("Can't convert fixed_size_list(list_size=%d) to matrix with %d cols",
+             (int)list_length, Rf_ncols(converter->ptype_view.ptype));
+  }
+
+  // First, we update the child array offset to account for the parent offset and
+  // materialize the child array.
+  child_converter->src.offset += raw_src_offset * list_length;
+  child_converter->src.length = child_length;
+  if (nanoarrow_converter_materialize_n(child_converter_xptr, child_length) !=
+      child_length) {
+    return EINVAL;
+  }
+
+  // If we have parent nulls, we have to project them into the destination
+  if (src->array_view->null_count != 0 &&
+      src->array_view->buffer_views[0].data.data != NULL) {
+    // Here, dst.offset has already been incremented such that it's at the end
+    // of the chunk, but we need the original one for fill_vec_with_nulls().
+    int64_t original_dst_offset = child_converter->dst.offset - child_length;
+    for (int64_t i = 0; i < src->length; i++) {
+      if (ArrowArrayViewIsNull(src->array_view, src->offset + i)) {
+        fill_vec_with_nulls(child_converter->dst.vec_sexp,
+                            original_dst_offset + (i * list_length), list_length);
+      }
+    }
+  }
+
+  return NANOARROW_OK;
+}
+
 static int nanoarrow_materialize_base(struct RConverter* converter, SEXP converter_xptr) {
   struct ArrayViewSlice* src = &converter->src;
   struct VectorSlice* dst = &converter->dst;
@@ -485,6 +695,8 @@ static int nanoarrow_materialize_base(struct RConverter* converter, SEXP convert
       return nanoarrow_materialize_blob(src, dst, options);
     case VECTOR_TYPE_LIST_OF:
       return nanoarrow_materialize_list_of(converter, converter_xptr);
+    case VECTOR_TYPE_MATRIX:
+      return nanoarrow_materialize_matrix(converter, converter_xptr);
     case VECTOR_TYPE_DATA_FRAME:
       return nanoarrow_materialize_data_frame(converter, converter_xptr);
     default:
